/* Inflate state-machine interface. */

/* Copyright (C) 2009 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <time.h>

#include "config.h"
#include "inflate.h"
#include "fixed-code.h"

static inline unsigned
min(unsigned x, unsigned y)
{
  return x < y ? x : y;
}

static inline unsigned
min3(unsigned x, unsigned y, unsigned z)
{
  if (z < x)
    x = z;
  return x < y ? x : y;
}

static void
PRINTF_STYLE(1, 2)  
werror(const char *format, ...)
{
  va_list args;
  va_start(args, format);
  vfprintf(stderr, format, args);
  va_end(args);
}

#define INFLATE_STATE_START 0
#define INFLATE_STATE_ERROR 1

void
inflate_init (struct inflate_ctx *self)
{
  self->status = INFLATE_FLAG_INPUT_EXPECTED;
  self->error = NULL;

  self->state = INFLATE_STATE_START;
  self->final = 0;
  self->input_buffer.n = self->input_buffer.bits = 0;
  self->output_end = self->output_size = 0;
}

static void
input_reset(struct inflate_ctx *self)
{
  self->input_buffer.n = self->input_buffer.bits = 0;
}

static unsigned
output_copy(struct inflate_ctx *self, unsigned length, const char *data)
{
  unsigned to_copy = min(length, INFLATE_WINDOW_MAX - self->output_end);
  memcpy(self->output_window + self->output_end, data, to_copy);

  self->output_end = (self->output_end + to_copy) & INFLATE_WINDOW_MASK;
  self->output_size += to_copy;
  if (self->output_size > INFLATE_WINDOW_MAX)
    self->output_size = INFLATE_WINDOW_MAX;

  self->length -= to_copy;
  return to_copy;
}

#define ERROR(s) do {		       \
    self->status = INFLATE_FLAG_ERROR; \
    self->error = (s);		       \
    self->state = INFLATE_STATE_ERROR; \
    goto yield_read;		       \
  } while (0)

#define GET_BITS(var, nbits) do {					\
  case __LINE__: self->state = __LINE__;				\
    while (self->input_buffer.n < (nbits))				\
      {									\
	if (consumed == input_length)					\
	  goto yield_read;						\
	self->input_buffer.bits						\
	  += (unsigned char) input[consumed++] << self->input_buffer.n;	\
	self->input_buffer.n += 8;					\
      }									\
    (var) = self->input_buffer.bits & ((1 << (nbits)) - 1);		\
    self->input_buffer.bits >>= (nbits);				\
    self->input_buffer.n -= (nbits);					\
  } while (0)

#define GET_CODE(decoder, value) do {					\
  int res;								\
 case __LINE__: self->state = __LINE__;					\
  do									\
    {									\
      if (self->input_buffer.n == 0)					\
	{								\
	  if (consumed == input_length)					\
	    goto yield_read;						\
									\
	  self->input_buffer.bits += (unsigned char) input[consumed++];	\
	  self->input_buffer.n = 8;					\
	}								\
									\
      res = huffman_decode_bits (decoder, &self->input_buffer, &value);	\
    } while (!res);							\
  if (res < 0)								\
    ERROR("Invalid huffman code");					\
} while (0)

/* If we're about to wrap around the end of the window, caller must
   first process the current output. */
#define OUTPUT_CHECK do {						\
  case __LINE__: self->state = __LINE__;				\
    if (produced > 0 && self->output_end == 0)				\
      goto yield_write;							\
  } while (0)

#define PUT_CHAR(c) do {						\
    self->output_window[self->output_end] = (c);			\
    self->output_end = (self->output_end + 1) & INFLATE_WINDOW_MASK;	\
    self->output_size += (self->output_size < INFLATE_WINDOW_MAX);	\
    produced++;								\
  } while (0)

/* NOTE: This function must avoid local variables. All state that is
   needed over a return, GET or OUTPUT_CHECK must be saved in the ctx
   struct.

   All writes to the output buffer are via the PUT_CHAR macro and the
   output_copy function. */
unsigned
inflate (struct inflate_ctx *self, unsigned input_length, const char *input,
	 unsigned *output_length, const char **output)
{
  unsigned consumed = 0;
  unsigned produced = 0;

  /* Set up for output via the output window. */
  *output = self->output_window + self->output_end;
  self->status &= ~INFLATE_FLAG_OUTPUT_PENDING;

  switch (self->state)
    {
    case INFLATE_STATE_ERROR:
    default:
      abort();

      while (!self->final)
	{
	  unsigned header;
	  unsigned type;

	case INFLATE_STATE_START:
	  self->state = INFLATE_STATE_START;

	  GET_BITS(header, 3);

	  self->final = header & 1;
	  type = header >> 1;

	  werror("Block of type %d, final = %d\n", type, self->final);

	  /* Can't use the switch statement... */
	  if (type == 3)
	    ERROR("Unsupported block type");

	  else if (type == 0)
	    {
	      /* No compression. */
	      unsigned nlength;
	      unsigned avail;

	      input_reset(self);
	      GET_BITS(self->length, 16);
	      GET_BITS(nlength, 16);
	      if (self->length + nlength != 0xffff)
		ERROR("Invalid length for uncompressed block");

	      OUTPUT_CHECK;
	      avail = input_length - consumed;
	      if (!avail)
		goto yield_read;

	      if (self->length > INFLATE_WINDOW_MAX
		  && !produced)
		{
		  /* Direct caller to copy from input buffer. */
		  produced = min(avail, self->length - INFLATE_WINDOW_MAX);

		  *output = input + consumed;
		  consumed += produced;
		  self->length -= produced;
		  goto yield_write;		  
		}
	      else
		{
		  unsigned copied
		    = output_copy(self, min(self->length, avail), input + consumed);

		  consumed += copied;
		  produced += copied;

		  if (self->length > 0)
		    goto yield_write;
		}
	    }
	  else {
	    if (type == 1)
	      {
		huffman_init (&self->length_active, fixed_length_tree);
		huffman_init (&self->distance_active, fixed_distance_tree);
	      }
	    else
	      {
		assert(type == 2);

		GET_BITS(self->hlit, 5);
		self->hlit += 257;
		if (self->hlit > INFLATE_HLIT_MAX)
		  ERROR("Invalid dynamic coding, hlit too large");

		GET_BITS(self->hdist, 5);
		self->hdist += 1;
		if (self->hdist > INFLATE_HDIST_MAX)
		  ERROR("Invalid dynamic coding, hdist too large");

		GET_BITS(self->hclen, 4);
		self->hclen += 4;
		assert(self->hclen <= INFLATE_HCLEN_MAX);

		werror("hlit = %d, hdist = %d, hclen = %d\n",
		       self->hlit, self->hdist, self->hclen);
		
		memset(self->clen, 0, INFLATE_HCLEN_MAX);
		for (self->k = 0; self->k < self->hclen; self->k++)
		  {
		    static const unsigned char permutation[INFLATE_HCLEN_MAX] =
		      { 16, 17, 18, 0, 8, 7, 9, 6, 10, 5,
			11, 4, 12, 3, 13, 2, 14, 1, 15 };

		    unsigned bits;
		    GET_BITS(bits, 3);
		    self->clen[permutation[self->k]] = bits;
		  }

		/* We reuse the length decoder for this. */
		if (!huffman_tree(self->length_tree, INFLATE_HCLEN_MAX, self->clen))
		  ERROR("Invalid dynamic coding, bad huffman coding coding");

		huffman_init (&self->length_active, self->length_tree);
	    
		for (self->k = 0; self->k < self->hlit + self->hdist; )
		  {
		    GET_CODE(&self->length_active, self->code);
		    if (self->code < 16)
		      self->clen[self->k++] = self->code;
		    else if (self->code == 16 && !self->k)
		      ERROR("Invalid dynamic coding, first code is a repeat");
		    else
		      {
			static unsigned char nbits_table[INFLATE_HCLEN_MAX - 16] = { 2, 3, 7 };
			static unsigned char repeat_table[INFLATE_HCLEN_MAX - 16] =  {3, 3, 11 };
			unsigned bits;
			unsigned repeat;
			unsigned value;

			GET_BITS(bits, nbits_table[self->code - 16]);
			repeat = repeat_table[self->code - 16] + bits;

			if (self->k + repeat > self->hlit + self->hdist)
			  ERROR("Invalid dynamic coding, repeat count too large");

			if (self->code == 16)
			  value = self->clen[self->k-1];
			else
			  value = 0;
			   
			memset(self->clen + self->k, value, repeat);
			self->k += repeat;
		      }
		  }
		if (!huffman_tree(self->length_tree, self->hlit, self->clen))
		  ERROR("Constructing length/literal huffman coding failed");
		huffman_init (&self->length_active, self->length_tree);

		if (self->hdist == 1 && self->clen[self->hlit] == 0)
		  {
		    werror("Dynamic block with literals only.");
		    /* No distance tree, only literals in block.
		       Indicate with a NULL tree. */
		    /* FIXME: Create test data for this feature. */
		    self->distance_active.root = NULL;
		  }
		else
		  {
		    if (!huffman_tree(self->distance_tree, self->hdist, self->clen + self->hlit))
		      ERROR("Constructing distance huffman coding failed");

		    huffman_init (&self->distance_active, self->distance_tree);
		  }
	      }
	    for (;;)
	      {
		GET_CODE(&self->length_active, self->code);

		/* NOTE: The invalid values 286 and 287 can occur only when using
		   the fixed coding, which is defined to include them. */
		if (self->code >= INFLATE_HLIT_MAX)
		  ERROR("Invalid length value");
		else if (self->code < 256)
		  {
		    OUTPUT_CHECK;
		    PUT_CHAR(self->code);
		  }
		else if (self->code == 256)
		  break;
		else
		  {
		    static const unsigned base_table[INFLATE_HLIT_MAX - 257] =
		      {
			3, 4, 5, 6, 7, 8, 9, 10,
			11, 13, 15, 17,
			19, 23, 27, 31,
			35, 43, 51, 59,
			67, 83, 99, 115,
			131, 163, 195, 227,
			258
		      };      
		    unsigned bits;
		    unsigned match;

		    if (!self->distance_active.root)
		      /* Only literals expected! */
		      ERROR("Length/distance pair in literal-only block");

		    if (self->code >= 265 && self->code < 285)
		      GET_BITS(bits, (self->code - 261) / 4);
		    else
		      bits = 0;

		    self->length = base_table[self->code - 257] + bits;

		    GET_CODE(&self->distance_active, self->code);
		    if (self->code >= 4)
		      {
			GET_BITS(bits, self->code / 2 - 1);
			self->code = ((2 + (self->code & 1)) << (self->code / 2 - 1)) + bits;
		      }
		    self->code += 1;

		    /* Length in self->length, distance in self->code. */
		    assert(self->length > 0);
		    assert(self->code > 0);
		    assert(self->code <= INFLATE_WINDOW_MAX);
		    
		    if (self->code > self->output_size)
		      ERROR("Invalid distance");

		    /* Note that length > distance is allowed. */		    
		    OUTPUT_CHECK;

		    match = (self->output_end - self->code) & INFLATE_WINDOW_MASK;

		    /* Check for common case: No overlap of source and
		       destination areas, and match not crossing the
		       window edge. */		    
		    if (self->length <= self->code
			&& match + self->length <= INFLATE_WINDOW_MAX
			&& (match + self->length <= self->output_end
			    || match >= self->output_end + self->length))
		      {
			produced
			  += output_copy(self, self->length,
					 self->output_window + match);
		      }
		    else
		      {
			do
			  {
			    PUT_CHAR(self->output_window[match]);
			    self->length--;
			    match = (match + 1) & INFLATE_WINDOW_MASK;
			  } while (self->length > 0 && self->output_end > 0);
		      }
		    if (self->length > 0)
		      goto yield_write;
		  }
	      }
	  }
	}
      
      self->state = INFLATE_STATE_ERROR;
      self->status &= ~INFLATE_FLAG_INPUT_EXPECTED;
      goto yield_read;
    }

 yield_write:
  self->status |= INFLATE_FLAG_OUTPUT_PENDING;
 yield_read:
  *output_length = produced;
  return consumed;
}
